import plotly.graph_objects as go

from plotly.offline import iplot
from warnings import warn

from echoviz.utils import BIN_CMAPS, HEAT_CMAPS

#TODO: 3D "annimation", add slider for frames
#TODO: gif making for sliced plot (3D would be hard to read)



def _frame_args(duration):
    return {"frame": {"duration": duration, "redraw": True},
            "mode": "immediate",
            "fromcurrent": True,
            "transition": {"duration": duration, "easing": "linear"}}


def animated_3d(vinputs, vlabels=None, vpreds=None, threshold=None,
                title="", show=True, filename=None):
    frames = []
    #TODO: Multithread this
    for i in range(len(vinputs)):
        data = [vinputs[i].make_mesh(opacity=0.3,
                                     showscale=False, colorscale="ice",
                                     reversescale=True)]
        if vlabels:
            for k in vlabels.keys():
                if vlabels[k][i].values.sum() == 0:
                    warn(f"{k.capitalize()} label for frame {i} is null,"
                         + " not plotting it.", RuntimeWarning)
                else:
                    data.append(vlabels[k][i].make_mesh(color=BIN_CMAPS[k][1]))
        if vpreds:
            for k in vpreds.keys():
                if vpreds[k][i].values.sum() == 0:
                    warn(f"{k.capitalize()} prediction for frame {i} is null,"
                         + " not plotting it.", RuntimeWarning)
                elif threshold:
                    data.append(vpreds[k][i].float2bool(threshold)
                                            .make_mesh(color=BIN_CMAPS[k][1]))
                else:
                    data.append(vpreds[k][i].make_mesh(colorscale=HEAT_CMAPS[k]))
        frames.append(go.Frame(data=data, name=str(i)))
    layout = {"sliders": [{
                "len": 0.8,
                'x': 0.1, 'y': 0,
                "steps": [{
                    "args": [[str(i)], _frame_args(0)],
                    "label": str(i),
                    "method": "animate",
                  } for i in range(len(frames))]}],
              "updatemenus": [{
                  "buttons": [{"args": [None, _frame_args(200)],
                               "label": "&#9654;", # Play symbol
                               "method": "animate"},
                              {"args": [[None], _frame_args(0)],
                               "label": "&#9724;", # Pause symbol
                               "method": "animate"}],
                  "direction": "left",
                  "type": "buttons",
                  'x': 0, 'y': 0
                }],
             }
    #FIXME: fix range to smooth animation
    fig = go.Figure(data=frames[0].data, frames=frames, layout=layout)
                    #range_x=[], range_y=[], range_z=[])
    camera = {"eye": {'x': -0.5, 'y': -1.15, 'z': 0.7},
              "up": {'x': 0, 'y': 0, 'z': 0.5}}
    fig.update_layout(scene_camera=camera, title=title, title_y=0.98,
                      plot_bgcolor="rgb(64,64,64)", autosize=False, width=600,
                      height=500, margin=dict(l=10, t=30, b=10))
    if show:
        iplot(fig)
    if filename:
        filename = filename.with_suffix(".html") if filename.suffix != ".html" else filename
        fig.write_html(filename) #FIXME: Reduce file size
    return frames



"""
Plot specific slice of 3D volume for better understanding of data
"""

import numpy as np
import plotly
import plotly.express as px
import plotly.graph_objects as go

from plotly.offline import iplot
from plotly.subplots import make_subplots

from echoviz.utils import BIN_CMAPS, HEAT_CMAPS



def _rot_slice(voxel_grid, key):
    return np.rot90(voxel_grid.values[key])


def plot_slice(vinput, vlabels, index, axis=1, vpreds=None, threshold=None,
               plot_input=True, title='', show=True, filename=None):
    # Expect vinput and vlabel to be VoxelGrid
    key = [slice(None), slice(None), slice(None)]
    key[axis] = index
    key = tuple(key)
    # Get them sliced
    sinput = _rot_slice(vinput, key)
    slabels = {k: _rot_slice(v, key).astype(np.uint8) for k, v in vlabels.items()}
    nb_col = (2 if vpreds else 1) + plot_input
    fig = make_subplots(cols=nb_col, horizontal_spacing=0.01)
    for c in range(nb_col):
        fig.add_trace(go.Heatmap(z=sinput, colorscale="Greys_r"), col=c+1, row=1)
    # Groundtruth
    for k in slabels.keys():
        fig.add_trace(go.Heatmap(z=slabels[k], colorscale=BIN_CMAPS[k]),
                      col=min(nb_col, 2), row=1)
    if vpreds:
        cmaps = BIN_CMAPS if threshold else HEAT_CMAPS
        for k in vpreds.keys():
            pred = np.rot90(vpreds[k].float2bool(threshold).values[key]) \
                    if threshold else _rot_slice(vpreds[k], key)
            fig.add_trace(go.Heatmap(z=pred, colorscale=cmaps[k]),
                          col=nb_col, row=1)
    fig.update_layout(title=title, autosize=False, margin=dict(l=10, r=10, t=30, b=10),
                      width=nb_col*350, height=350, title_y=0.98)
    fig.update_xaxes(showgrid=False, showticklabels=False)
    fig.update_yaxes(showgrid=False, showticklabels=False)
    fig.update_traces(showscale=False)
    if show:
        iplot(fig)
    if filename:
        #FIXME: Make file lighter
        fig.write_image(filename, format=filename.suffix.strip('.'))
    return fig



"""
3D plot of mitral valve voxels from voxel grids (see `utils/voxel_grid.py`)
Functions are adaptated from: https://www.raddq.com/dicom-processing-segmentation-visualization-in-python/
"""

import numpy as np
import plotly.graph_objects as go

from plotly.offline import iplot
from warnings import warn

from echoviz.utils import BIN_CMAPS, HEAT_CMAPS



def interactive_3d(vinput, vlabels=None, vpreds=None, threshold=None,
                   title="", show=True, filename=None):
    """
    """
    #FIXME? Add classes legend
    if vlabels and vpreds:
        warn("Plotting label and annotation in 3D make for a busy plot", UserWarning)
    camera = {"eye": {'x': -0.5, 'y': -1.15, 'z': 0.7},
              "up": {'x': 0, 'y': 0, 'z': 0.5}}
    inp_mesh = vinput.make_mesh(showscale=False,
                                opacity=0.3, colorscale="ice", reversescale=True)
    fig = go.Figure(data=inp_mesh)
    fig.update_layout(scene_camera=camera, title=title, title_y=0.98,
                      plot_bgcolor="rgb(64, 64, 64)",
                      autosize=False, width=600, height=500, margin=dict(l=10, t=30, b=10))
    if vlabels:
        lab_meshes = []
        for k in vlabels.keys():
            if vlabels[k].values.sum() == 0:
                warn(f"{k.capitalize()} label is null, not plotting it.",
                     RuntimeWarning)
            else:
                lab_meshes.append(vlabels[k].make_mesh(color=BIN_CMAPS[k][1]))
        fig.add_traces(lab_meshes)
    if vpreds:
        pred_meshes = []
        for k in vpreds.keys():
            if vpreds[k].values.sum() == 0:
                warn(f"{k.capitalize()} prediction is null, not plotting it.",
                     RuntimeWarning)
            elif threshold:
                pred_meshes.append(vpreds[k].float2bool(threshold)
                                            .make_mesh(color=BIN_CMAPS[k][1]))
            else:
                pred_meshes.append(vpreds[k].make_mesh(colorscale=HEAT_CMAPS[k]))
        fig.add_traces(pred_meshes)
    if show:
        iplot(fig)
    if filename:
        #FIXME: Diminish file size by removing js modules
        if filename.suffix != ".html":
            filename = filename.with_suffix(".html")
        fig.write_html(filename)
    return fig



import numpy as np
import plotly.graph_objects as go

from skimage.measure import marching_cubes



class VoxelGrid:
    # Code originating from Sverre Herland
    """ Wrapper for voxel grid and its relevant information for plotting """
    def __init__(self, grid, origin, directions, spacing):
        """
        grid:
        """
        grid = grid if isinstance(grid, np.ndarray) else np.array(grid)
        origin = origin if isinstance(origin, np.ndarray) else np.array(origin)
        directions = directions if isinstance(directions, np.ndarray) else np.array(directions)
        spacing = spacing if isinstance(spacing, np.ndarray) else np.array(spacing)
        assert grid.ndim == 3
        assert origin.shape == (3,)
        assert directions.shape == (3, 3)
        assert spacing.shape == (3,)
        self.values = grid
        self.origin = origin
        self.directions = directions
        self.spacing = spacing
        #FIXME: Allow to add devoxelize parameters as attributes

    @classmethod
    def fromh5(cls, filename):
        #TODO? Allow from open file too
        pass #TODO

    @classmethod
    def fromply(cls, filename, origin, directions, spacing):
        pass #TODO


    @property
    def shape(self):
        return self.values.shape

    def devoxelize(self, level=None, mask=False, stride=1):
        # Use for 3D plotting
        verts, faces, _, values = marching_cubes(self.values, level,
                                                 spacing=tuple(self.spacing),
                                                 step_size=stride, allow_degenerate=False,
                                                 mask=self.values > 0 if mask else None)
        # Map to coordinates given by voxel grid. Need right multiplication
        verts = verts @ self.directions + self.origin
        return verts, faces, values

    def make_mesh(self, level=None, mask=False, stride=1, **kwargs):
        # Create Plotly 3D mesh
        verts, faces, values = self.devoxelize(level, mask, stride)
        x, y, z = zip(*verts)
        i, j, k = zip(*faces)
        intensity = None if "color" in kwargs else values
        return go.Mesh3d(x=x, y=y, z=z, i=i, j=j, k=k, intensity=intensity, **kwargs)

    def float2bool(self, threshold=0.5, key=None):
        # Use for slice plotting
        #FIXME: Handle multiclass
        mask = (self.values[key] if key else self.values) > threshold
        return VoxelGrid(np.where(mask, 1, 0).astype(np.uint8), self.origin,
                         self.directions, self.spacing)



import h5py

from datetime import timedelta
from pathlib import Path
import subprocess
from time import sleep, time


from echoviz import (animated_3d, interactive_3d, static_3d,
                     plot_slice, sliced_sequence, sliced_volume,
                     VoxelGrid)




POUTS = {ext: Path(f"test-output.{ext}").resolve() for ext in ["png", "gif", "html"]}
YES = ['y', "yes", '']



def _time_plot(func, inputs, labels, *args, **kwargs):
    start = time()
    func(inputs, labels, *args, **kwargs)
    end = time()
    print(f"    `{func.__name__}` took {timedelta(seconds=(end - start))} to run.")

def _show_res(dtype, path):
    sentence = ("      Do you want to see the result from previous test as"
                f" {dtype}? (Y/n)")
    if input(sentence).lower() in YES:
        #FIXME: Don't assume firefox
        subprocess.call(["firefox", path.as_uri()])


def testy_testa():
    # Files to plot
    proot = Path("~/Documents/outputs/test-predictions")
    pinput = Path("~/Documents/data/bjørnar-hdf-isotropic/20150825110314_FED.h5").expanduser()
    pbpred = proot.joinpath("bin_20150825110314_FED.h5").expanduser()
    pmpred = proot.joinpath("multi_20150825110314_FED.h5").expanduser()

    # Additional info for voxel grid
    ih5 = h5py.File(pinput, 'r')
    spacing = ih5["VolumeGeometry"]["resolution"][()]
    origin = ih5["VolumeGeometry"]["origin"][()]
    directions = ih5["VolumeGeometry"]["directions"][()]
    fnb = ih5["VolumeGeometry"]["frameNumber"][()]
    ih5.close()
    pbh5, pmh5 = h5py.File(pbpred, 'r'), h5py.File(pmpred, 'r')
    inputs = []
    anteriors, posteriors, targets = [], [], [] # Annotations
    pralls, prants, prposts = [], [], []
    # Load frames in each category (input, labels, prediction)
    for i in range(1, fnb):
        # Input is identical in both files
        h5vol = pbh5["Input"][f"vol{i:02d}"][()]
        inputs.append(VoxelGrid(h5vol, origin, directions, spacing))
        # Binary segmentation target
        tg = pbh5["Target"][f"vol{i:02d}"][()]
        targets.append(VoxelGrid(tg, origin, directions, spacing))
        # Multi segmantation target
        ant = pmh5["Target"][f"anterior-{i:02d}"][()]
        anteriors.append(VoxelGrid(ant, origin, directions, spacing))
        post = pmh5["Target"][f"posterior-{i:02d}"][()]
        posteriors.append(VoxelGrid(post, origin, directions, spacing))
        # Predictions
        pra = pbh5["Prediction"][f"vol{i:02d}"][()]
        pralls.append(VoxelGrid(pra, origin, directions, spacing))
        prant = pmh5["Prediction"][f"anterior-{i:02d}"][()]
        prpost = pmh5["Prediction"][f"posterior-{i:02d}"][()]
        prants.append(VoxelGrid(prant, origin, directions, spacing))
        prposts.append(VoxelGrid(prpost, origin, directions, spacing))
    pbh5.close(), pmh5.close()
    # Try the plotting
    print("Here's some 2D static plots.")
    _time_plot(plot_slice, inputs[0],
               {"anterior": anteriors[0], "posterior": posteriors[0]},
               120, axis=1, plot_input=False, title="Labels on input")
    _time_plot(plot_slice, inputs[0], {"all": targets[0]},
               120, axis=2, vpreds={"all": pralls[0]}, threshold=0.5,
               filename=POUTS["png"], title="Labels on input & predictions")
    _show_res("image", POUTS["png"])
    print("Now for some 3D static plots.")
    _time_plot(static_3d, inputs[-1],
               {"anterior": anteriors[-1], "posterior": posteriors[-1]},
               vpreds={"anterior": prants[-1], "posterior": prposts[-1]},
               threshold=0.5, filename=POUTS["png"],
               title="Input, labels & predictions")
    _show_res("image", POUTS["png"])
    print("Let's make it fancy and have an interactive 3D plot")
    _time_plot(interactive_3d, inputs[-1],
               {"anterior": anteriors[-1], "posterior": posteriors[-1]},
               show=False, filename=POUTS["html"], title="Input & labels")
    _show_res("web file", POUTS["html"])
    print("Back to 2D! But with a twist... It's animated o/")
    _time_plot(sliced_sequence, inputs,
               {"anterior": anteriors, "posterior": posteriors},
               60, axis=0, filename=POUTS["gif"], title="Input & labels")
    _show_res("gif", POUTS["gif"])
    _time_plot(sliced_volume, inputs[-2], {"all": targets[-2]},
               axis=2, vpreds={"all": pralls[-2]}, plot_input=False,
               filename=POUTS["gif"], title="Labels on input & predictions")
    _show_res("gif", POUTS["gif"])
    print("And for the grand finale... Interactive 3D animated plot!!")
    _time_plot(animated_3d, inputs, None,
               vpreds={"anterior": prants, "posterior": prposts},
               filename=POUTS["html"], title="Input & predictions", show=False)
    _show_res("web file", POUTS["html"])
    sleep(3) # Wait so last file can be seen before being removed
    for p in POUTS.values(): #Remove created files
        p.unlink(missing_ok=True)



if __name__ == "__main__":
    testy_testa()



from echoviz.threed import animated_3d, interactive_3d, static_3d
from echoviz.twod import plot_slice, sliced_sequence, sliced_volume
from echoviz.utils import VoxelGrid



__author__ = "Maïlys Hau"
__version__ = "0.1.2"



import numpy as np
import plotly.express.colors as pxc



def rgb2rgba(rgb_cmap, min_transparency=0.1, max_transparency=0.7):
    """ Convert RGB scale to RGBA """
    a_cmap = np.linspace(min_transparency, max_transparency, len(rgb_cmap))
    rgba_cmap = []
    for c, a in zip(rgb_cmap, a_cmap):
        cl = c.strip("rgb()").split(',')
        rgba_cmap.append(f"rgba({cl[0]},{cl[1]},{cl[2]},{a})")
    return rgba_cmap



BIN_CMAPS = {"anterior": rgb2rgba(["rgb(145,53,125)"] * 2, 0),
             "posterior": rgb2rgba(["rgb(8,81,156)"] * 2, 0),
             "all": rgb2rgba(["rgb(153,52,4)"] * 2, 0)}
HEAT_CMAPS = {"anterior": rgb2rgba(pxc.sequential.Magenta),
              "posterior": rgb2rgba(pxc.sequential.Blues),
              "all": rgb2rgba(pxc.sequential.YlOrBr)}
